#include <string.h>
#include "erl_nif.h"
#include "concurrentqueue.h"

struct NifTraits : public moodycamel::ConcurrentQueueDefaultTraits
{
    // static const size_t BLOCK_SIZE = 8;
    // static const size_t EXPLICIT_INITIAL_INDEX_SIZE = 4;
    // static const size_t IMPLICIT_INITIAL_INDEX_SIZE = 4;
    // static const size_t INITIAL_IMPLCICIT_PRODUCER_HASH_SIZE = 1;
    // static const std::uint32_t EXPLICIT_CONSUMER_CONSUMPTION_QUOTA_BEFORE_ROTATE = 24;

    static inline void* malloc(std::size_t size) { return enif_alloc(size); }
    static inline void free(void* ptr) { enif_free(ptr); }
};

using lfqIns =  moodycamel::ConcurrentQueue<ErlNifBinary, NifTraits> *;

ERL_NIF_TERM atomOk;
ERL_NIF_TERM atomError;
ERL_NIF_TERM atomTrue;
ERL_NIF_TERM atomFalse;
ERL_NIF_TERM atomEmpty;

ERL_NIF_TERM make_binary(ErlNifEnv *env, const char *buff, size_t length) {
    ERL_NIF_TERM term;
    unsigned char *destination_buffer = enif_make_new_binary(env, length, &term);
    memcpy(destination_buffer, buff, length);
    return term;
}

ERL_NIF_TERM make_error(ErlNifEnv *env, const char *error) {
    return enif_make_tuple2(env, atomError, make_binary(env, error, strlen(error)));
}

void eLfqFree(ErlNifEnv *, void *obj) {
    lfqIns *ObjIns = static_cast<lfqIns *>(obj);
    if (ObjIns != nullptr) {
        ErlNifBinary TermBin;
        while ((*ObjIns)->try_dequeue(TermBin)) {
            enif_release_binary(&TermBin);
        }

        delete ObjIns;
    }
}

int nifLoad(ErlNifEnv *env, void **priv_data, ERL_NIF_TERM) {
    ErlNifResourceFlags flags = static_cast<ErlNifResourceFlags>(ERL_NIF_RT_CREATE | ERL_NIF_RT_TAKEOVER);
    ErlNifResourceType *ResIns = NULL;
    ResIns = enif_open_resource_type(env, NULL, "eLfqRes", eLfqFree, flags, NULL);
    if (NULL == ResIns)
        return -1;

    *priv_data = ResIns;
    atomOk = enif_make_atom(env, "ok");
    atomError = enif_make_atom(env, "lfq_error");
    atomTrue = enif_make_atom(env, "true");
    atomFalse = enif_make_atom(env, "false");
    atomEmpty = enif_make_atom(env, "lfq_empty");

    return 0;
}

int nifUpgrade(ErlNifEnv *env, void **priv_data, void **, ERL_NIF_TERM) {
    ErlNifResourceFlags flags = static_cast<ErlNifResourceFlags>(ERL_NIF_RT_CREATE | ERL_NIF_RT_TAKEOVER);
    ErlNifResourceType *ResIns = NULL;
    ResIns = enif_open_resource_type(env, NULL, "eLfqRes", eLfqFree, flags, NULL);
    if (NULL == ResIns)
        return -1;

    *priv_data = ResIns;
    return 0;
}

void nifUnload(ErlNifEnv *, void *priv_data) {
}

ERL_NIF_TERM nifNew(ErlNifEnv *env, int, const ERL_NIF_TERM *) {
    ErlNifResourceType *ResIns = static_cast<ErlNifResourceType *>(enif_priv_data(env));

    lfqIns* ObjIns = static_cast<lfqIns *>(enif_alloc_resource(ResIns, sizeof(lfqIns)));
    *ObjIns = new moodycamel::ConcurrentQueue<ErlNifBinary, NifTraits>;

    if (ObjIns == NULL)
        return make_error(env, "enif_alloc_resource failed");

    ERL_NIF_TERM RefTerm = enif_make_resource(env, ObjIns);
    enif_release_resource(ObjIns);
    return enif_make_tuple2(env, atomOk, RefTerm);
}

ERL_NIF_TERM nifIn2(ErlNifEnv *env, int, const ERL_NIF_TERM argv[]) {
    ErlNifResourceType *ResIns = static_cast<ErlNifResourceType *>(enif_priv_data(env));

    lfqIns *ObjIns = NULL;

    if (!enif_get_resource(env, argv[0], ResIns, (void **) &ObjIns)) {
        return enif_make_badarg(env);
    }

    ErlNifBinary TermBin;
    if(!enif_term_to_binary(env, argv[1], &TermBin))
        return enif_make_badarg(env);

    if ((*ObjIns)->enqueue(TermBin)){
        return atomTrue;
    } else {
        enif_release_binary(&TermBin);
        return atomFalse;
    }
}

ERL_NIF_TERM nifIns2(ErlNifEnv *env, int, const ERL_NIF_TERM argv[]) {
    ErlNifResourceType *ResIns = static_cast<ErlNifResourceType *>(enif_priv_data(env));

    lfqIns *ObjIns = NULL;

    if (!enif_get_resource(env, argv[0], ResIns, (void **) &ObjIns)) {
        return enif_make_badarg(env);
    }

    std::vector<ErlNifBinary> data;
    for (counter_t i = 0; i != BULK_BATCH_SIZE; ++i) {
        data.push_back(i);
    }

    ErlNifBinary TermBin;
    if(!enif_term_to_binary(env, argv[1], &TermBin))
        return enif_make_badarg(env);

    if ((*ObjIns)->enqueue_bulk(data.cbegin(), data.size())){
        return atomTrue;
    } else {
        for(auto it = data.cbegin(), it != data.cend(), it++){
            enif_release_binary(&TermBin);
        }
        return atomFalse;
    }
}

ERL_NIF_TERM nifTryOut1(ErlNifEnv *env, int, const ERL_NIF_TERM argv[]) {
    ErlNifResourceType *ResIns = static_cast<ErlNifResourceType *>(enif_priv_data(env));
    lfqIns *ObjIns = NULL;

    if (!enif_get_resource(env, argv[0], ResIns, (void **) &ObjIns)) {
        return enif_make_badarg(env);
    }

    ErlNifBinary TermBin;

    if ((*ObjIns)->try_dequeue(TermBin)) {
        ERL_NIF_TERM OutTerm;
        if(enif_binary_to_term(env, TermBin.data, TermBin.size, &OutTerm, 0) == 0)
        {
            enif_release_binary(&TermBin);
            // return make_error(env, atomError "failed to decode data");
            return atomError;
        } else{
            enif_release_binary(&TermBin);
            return OutTerm;
        }
    } else {
        return atomEmpty;
    }
}

ERL_NIF_TERM nifTryOuts2(ErlNifEnv *env, int, const ERL_NIF_TERM argv[]) {
    return atomEmpty;
}


ERL_NIF_TERM nifSize1(ErlNifEnv *env, int, const ERL_NIF_TERM argv[]) {
    ErlNifResourceType *ResIns = static_cast<ErlNifResourceType *>(enif_priv_data(env));
    lfqIns *ObjIns = NULL;

    if (!enif_get_resource(env, argv[0], ResIns, (void **) &ObjIns)) {
        return enif_make_badarg(env);
    }

    size_t LfqSize = (*ObjIns)->size_approx();
    return enif_make_long(env, (long int)LfqSize);
}

static ErlNifFunc nifFuncs[] =
        {
                {"new",           0, nifNew},
                {"in",            2, nifIn2},
                {"ins",           2, nifIns2},
                {"tryOut",        1, nifTryOut1},
                {"tryOuts",       2, nifTryOuts2},
                {"size",          1, nifSize1}
        };

ERL_NIF_INIT(eLfq, nifFuncs, nifLoad, NULL, nifUpgrade, nifUnload)